from .callback import (Callback, WriteDataCallback, WriteFilesCallback, WriteNamesCallback, WriteSubjectCallback,
                       WriteImageInformationCallback, ComposeCallback, get_default_callbacks)
from .fileloader import (Load, LoadDefault)
from .writer import (Hdf5Writer, Writer, get_writer)
from .traverser import (SubjectFileTraverser, Traverser)



import abc
import pickle

import numpy as np

from . import transformation as tfm


class Assembler(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def add_batch(self, prediction, batch: dict):
        pass


def numpy_zeros(shape: tuple, id_: str):
    return np.zeros(shape)


class SubjectAssembler(Assembler):
    """Assembles predictions of one or multiple subjects.

    Assumes that the network output, i.e. to_assemble, is of shape (B, ..., C)
    where B is the batch size and C is the numbers of channels (must be at least 1). ... refers to the arbitrary image
    dimension.
    """

    def __init__(self, zero_fn=numpy_zeros):
        self.predictions = {}
        self.subjects_ready = set()
        self.zero_fn = zero_fn

    def add_batch(self, to_assemble, batch: dict, last_batch=False, transform: tfm.Transform=None):
        # todo(fabianbalsiger): replace tfm.Transform by callback
        if 'subject_index' not in batch:
            raise ValueError('SubjectAssembler requires "subject_index" to be extracted (use IndexingExtractor)')
        if 'index_expr' not in batch:
            raise ValueError('SubjectAssembler requires "index_expr" to be extracted (use IndexingExtractor)')
        if 'shape' not in batch:
            raise ValueError('SubjectAssembler requires "shape" to be extracted (use ImageShapeExtractor)')

        if not isinstance(to_assemble, dict):
            to_assemble = {'__prediction': to_assemble}

        for idx in range(len(batch['subject_index'])):
            self.add_sample(to_assemble, batch, idx, transform)

        if last_batch:
            # to prevent from last batch to be ignored
            self.end()

    def end(self):
        self.subjects_ready = set(self.predictions.keys())

    def add_sample(self, to_assemble, batch, idx, transform: tfm.Transform=None):
        # todo(fabianbalsiger): replace tfm.Transform by callback
        # initialize subject
        subject_index = batch['subject_index'][idx]

        if subject_index not in self.predictions and not self.predictions:
            self.predictions[subject_index] = self._init_new_subject(batch, to_assemble, idx)
        elif subject_index not in self.predictions:
            self.subjects_ready = set(self.predictions.keys())
            self.predictions[subject_index] = self._init_new_subject(batch, to_assemble, idx)

        for key in to_assemble:
            data = to_assemble[key][idx]
            if transform is not None:
                data = transform({key: data, 'batch': batch, 'batch_idx': idx})[key]

            index_expr = batch['index_expr'][idx]
            if isinstance(index_expr, bytes):
                # is pickled
                index_expr = pickle.loads(index_expr)

            self.predictions[subject_index][key][index_expr.expression] = data

    def _init_new_subject(self, batch, to_assemble, idx):
        subject_prediction = {}
        for key in to_assemble:
            subject_shape = batch['shape'][idx]
            subject_shape += (to_assemble[key].shape[-1],)
            subject_prediction[key] = self.zero_fn(subject_shape, key)
        return subject_prediction

    def get_assembled_subject(self, subject_index: int):
        """Gets the prediction of a subject.

        Args:
            subject_index (int): The subject's index.

        Returns:
            np.ndarray: The prediction of the subject.
        """
        try:
            self.subjects_ready.remove(subject_index)
        except KeyError:
            # check if subject is assembled but not listed as ready
            # this can happen if only one subject was assembled or last
            if subject_index not in self.predictions:
                raise ValueError('Subject with index {} not in assembler'.format(subject_index))
        assembled = self.predictions.pop(subject_index)
        if '__prediction' in assembled:
            return assembled['__prediction']
        return assembled




import abc

import numpy as np


class ConfusionMatrix:
    """Represents a confusion matrix (or error matrix)."""

    def __init__(self, prediction, label):
        """Initializes a new instance of the ConfusionMatrix class."""

        # true positive (tp): we predict a label of 1 (positive), and the true label is 1
        self.tp = np.sum(np.logical_and(prediction == 1, label == 1))
        # true negative (tn): we predict a label of 0 (negative), and the true label is 0
        self.tn = np.sum(np.logical_and(prediction == 0, label == 0))
        # false positive (fp): we predict a label of 1 (positive), but the true label is 0
        self.fp = np.sum(np.logical_and(prediction == 1, label == 0))
        # false negative (fn): we predict a label of 0 (negative), but the true label is 1
        self.fn = np.sum(np.logical_and(prediction == 0, label == 1))

        self.n = prediction.size


class IMetric(metaclass=abc.ABCMeta):
    """Represents an evaluation metric."""

    def __init__(self):
        self.metric = 'IMetric'

    @abc.abstractmethod
    def calculate(self):
        """Calculates the metric."""

        raise NotImplementedError

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return '{self.metric}' \
            .format(self=self)


class IConfusionMatrixMetric(IMetric):
    """Represents an evaluation metric based on the confusion matrix."""

    def __init__(self):
        """Initializes a new instance of the IConfusionMatrixMetric class."""
        super().__init__()
        self.metric = 'IConfusionMatrixMetric'
        self.confusion_matrix = None  # ConfusionMatrix

    @abc.abstractmethod
    def calculate(self):
        """Calculates the metric."""

        raise NotImplementedError


class ISimpleITKImageMetric(IMetric):
    """Represents an evaluation metric based on SimpleITK images."""

    def __init__(self):
        """Initializes a new instance of the ISimpleITKImageMetric class."""
        super(ISimpleITKImageMetric, self).__init__()
        self.metric = 'ISimpleITKImageMetric'
        self.ground_truth = None  # SimpleITK.Image
        self.segmentation = None  # SimpleITK.Image

    @abc.abstractmethod
    def calculate(self):
        """Calculates the metric."""

        raise NotImplementedError


class INumpyArrayMetric(IMetric):
    """Represents an evaluation metric based on numpy arrays."""

    def __init__(self):
        """Initializes a new instance of the INumpyArrayMetric class."""
        super(INumpyArrayMetric, self).__init__()
        self.metric = 'INumpyArrayMetric'
        self.ground_truth = None  # np.ndarray
        self.segmentation = None  # np.ndarray

    @abc.abstractmethod
    def calculate(self):
        """Calculates the metric."""

        raise NotImplementedError


class Information(IMetric):
    """Represents an information.

    Can be used to add an additional column of information to an evaluator.
    """

    def __init__(self, column_name: str, value: str):
        """Initializes a new instance of the Information class."""
        super().__init__()
        self.metric = column_name
        self.value = value

    def calculate(self):
        """Outputs the value of the information."""
        return self.value
